#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import typing
import boto3
from apache_beam.options import pipeline_options

from firehose_pyio.options import FirehoseOptions

__all__ = ["FirehoseClient", "FirehoseClientError"]


def get_http_error_code(exc):
    if hasattr(exc, "response"):
        return exc.response.get("ResponseMetadata", {}).get("HTTPStatusCode")
    return None


class FirehoseClientError(Exception):
    def __init__(self, message=None, code=None):
        self.message = message
        self.code = code


class FirehoseClient(object):
    """
    Wrapper for boto3 library.
    """

    def __init__(self, options: typing.Union[FirehoseOptions, dict]):
        """Constructor of the FirehoseClient.

        Args:
            options (Union[FirehoseOptions, dict]): Options to create a boto3 Firehose client.
        """
        assert boto3 is not None, "Missing boto3 requirement"
        if isinstance(options, pipeline_options.PipelineOptions):
            options = options.view_as(FirehoseOptions)
            access_key_id = options.aws_access_key_id
            secret_access_key = options.aws_secret_access_key
            session_token = options.aws_session_token
            endpoint_url = options.endpoint_url
            use_ssl = not options.disable_ssl
            region_name = options.region_name
            api_version = options.api_version
            verify = options.verify
        else:
            access_key_id = options.get("aws_access_key_id")
            secret_access_key = options.get("aws_secret_access_key")
            session_token = options.get("aws_session_token")
            endpoint_url = options.get("endpoint_url")
            use_ssl = not options.get("disable_ssl", False)
            region_name = options.get("region_name")
            api_version = options.get("api_version")
            verify = options.get("verify")

        session = boto3.session.Session()
        self.client = session.client(
            service_name="firehose",
            region_name=region_name,
            api_version=api_version,
            use_ssl=use_ssl,
            verify=verify,
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token,
        )

    def is_delivery_stream_active(self, delivery_stream_name: str):
        """Check if an Amazon Firehose delivery stream is active.

        Args:
            delivery_stream_name (str): Amazon Firehose delivery stream name.

        Raises:
            FirehoseClientError: Firehose client error.

        Returns:
            (bool): Whether or not the given Firehose delivery stream is active.
        """
        try:
            boto_response = self.client.describe_delivery_stream(
                DeliveryStreamName=delivery_stream_name
            )
            return (
                boto_response["DeliveryStreamDescription"]["DeliveryStreamStatus"]
                == "ACTIVE"
            )
        except Exception as e:
            raise FirehoseClientError(str(e), get_http_error_code(e))

    def put_record_batch(
        self,
        records: list,
        delivery_stream_name: str,
        jsonify: bool,
        multiline: bool,
    ):
        """Put records to an Amazon Firehose delivery stream in batch.

        Args:
            records (list): Records to put into a Firehose delivery stream.
            delivery_stream_name (str): Amazon Firehose delivery stream name.
            jsonify (bool): Whether to convert records into JSON.
            multiline (bool): Whether to add a new line at the end of each record.

        Raises:
            FirehoseClientError: Firehose client error.
            TypeError: TypeError.

        Returns:
            (Object): Boto3 response message.
        """

        def process_data(record: str, jsonify: bool, multiline: bool):
            if jsonify:
                record = json.dumps(record)
            if multiline:
                record = f"{record}\n"
            return record

        if not isinstance(records, list):
            raise TypeError("Records should be a list.")
        try:
            boto_response = self.client.put_record_batch(
                DeliveryStreamName=delivery_stream_name,
                Records=[
                    {
                        "Data": process_data(
                            record=r, jsonify=jsonify, multiline=multiline
                        )
                    }
                    for r in records
                ],
            )
            return boto_response
        except Exception as e:
            raise FirehoseClientError(str(e), get_http_error_code(e))

    def close(self):
        """Closes underlying endpoint connections."""
        self.client.close()


class FakeFirehoseClient:
    def __init__(self, fake_config: dict):
        self.num_success = fake_config.get("num_success", 0)

    def put_record_batch(
        self, records: list, delivery_stream_name: str, jsonify: bool, multiline: bool
    ):
        if not isinstance(records, list):
            raise TypeError("Records should be a list.")
        request_responses = []
        for index, _ in enumerate(records):
            if index < self.num_success:
                request_responses.append({"RecordId": index})
            else:
                request_responses.append(
                    {"ErrorCode": "Error", "ErrorMessage": "This error"}
                )
        return {
            "FailedPutCount": ["RecordId" in r for r in request_responses],
            "Encrypted": False,
            "RequestResponses": request_responses,
        }

    def close(self):
        return



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing
import apache_beam as beam
from apache_beam import metrics
from apache_beam.pvalue import PCollection


from firehose_pyio.boto3_client import FirehoseClient, FakeFirehoseClient
from firehose_pyio.options import FirehoseOptions

__all__ = ["WriteToFirehose"]


class _FirehoseWriteFn(beam.DoFn):
    """Create the connector can put records in batch to an Amazon Firehose delivery stream.

    Args:
        delivery_stream_name (str): Amazon Firehose delivery stream name.
        jsonify (bool): Whether to convert records into JSON.
        multiline (bool): Whether to add a new line at the end of each record.
        max_trials (int): Maximum number of trials to put failed records.
        options (Union[FirehoseOptions, dict]): Options to create a boto3 Firehose client.
        fake_config (dict, optional): Config parameters when using FakeFirehoseClient for testing.
    """

    total_elements_count = metrics.Metrics.counter(
        "_FirehoseWriteFn", "total_elements_count"
    )
    succeeded_elements_count = metrics.Metrics.counter(
        "_FirehoseWriteFn", "succeeded_elements_count"
    )
    failed_elements_count = metrics.Metrics.counter(
        "_FirehoseWriteFn", "failed_elements_count"
    )

    def __init__(
        self,
        delivery_stream_name: str,
        jsonify: bool,
        multiline: bool,
        max_trials: int,
        options: typing.Union[FirehoseOptions, dict],
        fake_config: dict,
    ):
        """Constructor of _FirehoseWriteFn

        Args:
            delivery_stream_name (str): Amazon Firehose delivery stream name.
            jsonify (bool): Whether to convert records into JSON.
            multiline (bool): Whether to add a new line at the end of each record.
            max_trials (int): Maximum number of trials to put failed records.
            options (Union[FirehoseOptions, dict]): Options to create a boto3 Firehose client.
            fake_config (dict, optional): Config parameters when using FakeFirehoseClient for testing.
        """
        super().__init__()
        self.delivery_stream_name = delivery_stream_name
        self.jsonify = jsonify
        self.multiline = multiline
        self.max_trials = max_trials
        self.options = options
        self.fake_config = fake_config

    def start_bundle(self):
        if not self.fake_config:
            self.client = FirehoseClient(self.options)
            assert (
                self.client.is_delivery_stream_active(self.delivery_stream_name) is True
            )
        else:
            self.client = FakeFirehoseClient(self.fake_config)

    def process(self, element):
        if isinstance(element, tuple):
            element = element[1]
        loop, total, failed = 0, len(element), []
        while loop < self.max_trials:
            responses = self.client.put_record_batch(
                element, self.delivery_stream_name, self.jsonify, self.multiline
            )["RequestResponses"]
            for index, result in enumerate(responses):
                if "RecordId" not in result:
                    failed.append(element[index])
            if len(failed) == 0 or (self.max_trials - loop == 1):
                break
            element = failed
            failed = []
            loop += 1
        self.total_elements_count.inc(total)
        self.succeeded_elements_count.inc(total - len(failed))
        self.failed_elements_count.inc(len(failed))
        return failed

    def finish_bundle(self):
        self.client.close()


class WriteToFirehose(beam.PTransform):
    """A transform that puts records into an Amazon Firehose delivery stream

    Takes an input PCollection and put them in batch using the boto3 package.
    For more information, visit the `Boto3 Documentation <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/firehose/client/put_record_batch.html>`__.

    Note that, if the PCollection element is a tuple (i.e. keyed stream), only the value is used to put records in batch.

    Args:
        delivery_stream_name (str): Amazon Firehose delivery stream name.
        jsonify (bool): Whether to convert records into JSON.
        multiline (bool): Whether to add a new line at the end of each record.
        max_trials (int): Maximum number of trials to put failed records. Defaults to 3.
        fake_config (dict, optional): Config parameters when using FakeFirehoseClient for testing. Defaults to {}.
    """

    def __init__(
        self,
        delivery_stream_name: str,
        jsonify: bool,
        multiline: bool,
        max_trials: int = 3,
        fake_config: dict = {},
    ):
        """Constructor of the transform that puts records into an Amazon Firehose delivery stream

        Args:
            delivery_stream_name (str): Amazon Firehose delivery stream name.
            jsonify (bool): Whether to convert records into JSON.
            multiline (bool): Whether to add a new line at the end of each record.
            max_trials (int): Maximum number of trials to put failed records. Defaults to 3.
            fake_config (dict, optional): Config parameters when using FakeFirehoseClient for testing. Defaults to {}.
        """
        super().__init__()
        self.delivery_stream_name = delivery_stream_name
        self.jsonify = jsonify
        self.multiline = multiline
        self.max_trials = max_trials
        self.fake_config = fake_config

    def expand(self, pcoll: PCollection):
        options = pcoll.pipeline.options.view_as(FirehoseOptions)
        return pcoll | beam.ParDo(
            _FirehoseWriteFn(
                self.delivery_stream_name,
                self.jsonify,
                self.multiline,
                self.max_trials,
                options,
                self.fake_config,
            )
        )



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from apache_beam.options.pipeline_options import PipelineOptions


class FirehoseOptions(PipelineOptions):
    @classmethod
    def _add_argparse_args(cls, parser):
        # These options are passed to the Firehose PyIO Client
        parser.add_argument(
            "--aws_access_key_id",
            default=None,
            help="The secret key to use when creating the boto3 client.",
        )
        parser.add_argument(
            "--aws_secret_access_key",
            default=None,
            help="The secret key to use when creating the boto3 client.",
        )
        parser.add_argument(
            "--aws_session_token",
            default=None,
            help="The session token to use when creating the boto3 client.",
        )
        parser.add_argument(
            "--endpoint_url",
            default=None,
            help="The complete URL to use for the constructed boto3 client.",
        )
        parser.add_argument(
            "--region_name",
            default=None,
            help="The name of the region associated with the boto3 client.",
        )
        parser.add_argument(
            "--api_version",
            default=None,
            help="The API version to use with the boto3 client.",
        )
        parser.add_argument(
            "--verify",
            default=None,
            help="Whether or not to verify SSL certificates with the boto3 client.",
        )
        parser.add_argument(
            "--disable_ssl",
            default=False,
            action="store_true",
            help=(
                "Whether or not to use SSL with the boto3 client. "
                "By default, SSL is used."
            ),
        )
